@@ -6,6 +6,7 @@ import os
 import sys
 import time
 from contextlib import contextmanager
+from itertools import combinations
 
 from nose import SkipTest
 from tornado.web import HTTPError
@@ -46,7 +47,7 @@ class TestFileContentsManager(TestCase):
 
     def symlink(self, contents_manager, src, dst):
         """Make a symlink to src from dst
-        
+
         src and dst are api_paths
         """
         src_os_path = contents_manager._get_os_path(src)
@@ -106,8 +107,8 @@ class TestFileContentsManager(TestCase):
         self.assertNotEqual(cp_dir, cp_subdir)
         self.assertEqual(cp_dir, os.path.join(root, cpm.checkpoint_dir, cp_name))
         self.assertEqual(cp_subdir, os.path.join(root, subd, cpm.checkpoint_dir, cp_name))
-    
-    @dec.skip_win32
+
+    @dec.skipif(sys.platform == 'win32' and sys.version_info[0] < 3)
     def test_bad_symlink(self):
         with TemporaryDirectory() as td:
             cm = FileContentsManager(root_dir=td)
@@ -119,9 +120,36 @@ class TestFileContentsManager(TestCase):
             # create a broken symlink
             self.symlink(cm, "target", '%s/%s' % (path, 'bad symlink'))
             model = cm.get(path)
-            self.assertEqual(model['content'], [file_model])
-    
-    @dec.skip_win32
+
+            contents = {
+                content['name']: content for content in model['content']
+            }
+            self.assertTrue('untitled.txt' in contents)
+            self.assertEqual(contents['untitled.txt'], file_model)
+            # broken symlinks should still be shown in the contents manager
+            self.assertTrue('bad symlink' in contents)
+
+    @dec.skipif(sys.platform == 'win32')
+    def test_recursive_symlink(self):
+        with TemporaryDirectory() as td:
+            cm = FileContentsManager(root_dir=td)
+            path = 'test recursive symlink'
+            _make_dir(cm, path)
+            os_path = cm._get_os_path(path)
+            os.symlink("recursive", os.path.join(os_path, "recursive"))
+            file_model = cm.new_untitled(path=path, ext='.txt')
+
+            model = cm.get(path)
+
+            contents = {
+                content['name']: content for content in model['content']
+            }
+            self.assertIn('untitled.txt', contents)
+            self.assertEqual(contents['untitled.txt'], file_model)
+            # recursive symlinks should not be shown in the contents manager
+            self.assertNotIn('recursive', contents)
+
+    @dec.skipif(sys.platform == 'win32' and sys.version_info[0] < 3)
     def test_good_symlink(self):
         with TemporaryDirectory() as td:
             cm = FileContentsManager(root_dir=td)
@@ -140,19 +168,19 @@ class TestFileContentsManager(TestCase):
                 sorted(dir_model['content'], key=lambda x: x['name']),
                 [symlink_model, file_model],
             )
-    
+
     def test_403(self):
         if hasattr(os, 'getuid'):
             if os.getuid() == 0:
                 raise SkipTest("Can't test permissions as root")
         if sys.platform.startswith('win'):
             raise SkipTest("Can't test permissions on Windows")
-        
+
         with TemporaryDirectory() as td:
             cm = FileContentsManager(root_dir=td)
             model = cm.new_untitled(type='file')
             os_path = cm._get_os_path(model['path'])
-        
+
             os.chmod(os_path, 0o400)
             try:
                 with cm.open(os_path, 'w') as f:
@@ -188,7 +216,42 @@ class TestFileContentsManager(TestCase):
 
 
 class TestContentsManager(TestCase):
-    
+    @contextmanager
+    def assertRaisesHTTPError(self, status, msg=None):
+        msg = msg or "Should have raised HTTPError(%i)" % status
+        try:
+            yield
+        except HTTPError as e:
+            self.assertEqual(e.status_code, status)
+        else:
+            self.fail(msg)
+
+    def make_populated_dir(self, api_path):
+        cm = self.contents_manager
+
+        self.make_dir(api_path)
+
+        cm.new(path="/".join([api_path, "nb.ipynb"]))
+        cm.new(path="/".join([api_path, "file.txt"]))
+
+    def check_populated_dir_files(self, api_path):
+        dir_model = self.contents_manager.get(api_path)
+
+        self.assertEqual(dir_model['path'], api_path)
+        self.assertEqual(dir_model['type'], "directory")
+
+        for entry in dir_model['content']:
+            if entry['type'] == "directory":
+                continue
+            elif entry['type'] == "file":
+                self.assertEqual(entry['name'], "file.txt")
+                complete_path = "/".join([api_path, "file.txt"])
+                self.assertEqual(entry["path"], complete_path)
+            elif entry['type'] == "notebook":
+                self.assertEqual(entry['name'], "nb.ipynb")
+                complete_path = "/".join([api_path, "nb.ipynb"])
+                self.assertEqual(entry["path"], complete_path)
+
     def setUp(self):
         self._temp_dir = TemporaryDirectory()
         self.td = self._temp_dir.name
@@ -198,10 +261,10 @@ class TestContentsManager(TestCase):
 
     def tearDown(self):
         self._temp_dir.cleanup()
-    
+
     def make_dir(self, api_path):
         """make a subdirectory at api_path
-        
+
         override in subclasses if contents are not on the filesystem.
         """
         _make_dir(self.contents_manager, api_path)
@@ -247,7 +310,7 @@ class TestContentsManager(TestCase):
         self.assertEqual(model['name'], 'Untitled Folder')
         self.assertEqual(model['path'], 'Untitled Folder')
         sub_dir = model['path']
-        
+
         model = cm.new_untitled(path=sub_dir)
         assert isinstance(model, dict)
         self.assertIn('name', model)
@@ -257,6 +320,12 @@ class TestContentsManager(TestCase):
         self.assertEqual(model['name'], 'untitled')
         self.assertEqual(model['path'], '%s/untitled' % sub_dir)
 
+        # Test with a compound extension
+        model = cm.new_untitled(path=sub_dir, ext='.foo.bar')
+        self.assertEqual(model['name'], 'untitled.foo.bar')
+        model = cm.new_untitled(path=sub_dir, ext='.foo.bar')
+        self.assertEqual(model['name'], 'untitled1.foo.bar')
+
     def test_modified_date(self):
 
         cm = self.contents_manager
@@ -460,6 +529,60 @@ class TestContentsManager(TestCase):
         # Check that a 'get' on the deleted notebook raises and error
         self.assertRaises(HTTPError, cm.get, path)
 
+    def test_rename(self):
+        cm = self.contents_manager
+        # Create a new notebook
+        nb, name, path = self.new_notebook()
+
+        # Rename the notebook
+        cm.rename(path, "changed_path")
+
+        # Attempting to get the notebook under the old name raises an error
+        self.assertRaises(HTTPError, cm.get, path)
+        # Fetching the notebook under the new name is successful
+        assert isinstance(cm.get("changed_path"), dict)
+
+        # Test validation.  Currently, only Windows has a non-empty set of invalid characters
+        if sys.platform == 'win32' and isinstance(cm, FileContentsManager):
+            with self.assertRaisesHTTPError(400):
+                cm.rename("changed_path", "prevent: in name")
+
+        # Ported tests on nested directory renaming from pgcontents
+        all_dirs = ['foo', 'bar', 'foo/bar', 'foo/bar/foo', 'foo/bar/foo/bar']
+        unchanged_dirs = all_dirs[:2]
+        changed_dirs = all_dirs[2:]
+
+        for _dir in all_dirs:
+            self.make_populated_dir(_dir)
+            self.check_populated_dir_files(_dir)
+
+        # Renaming to an existing directory should fail
+        for src, dest in combinations(all_dirs, 2):
+            with self.assertRaisesHTTPError(409):
+                cm.rename(src, dest)
+
+        # Creating a notebook in a non_existant directory should fail
+        with self.assertRaisesHTTPError(404):
+            cm.new_untitled("foo/bar_diff", ext=".ipynb")
+
+        cm.rename("foo/bar", "foo/bar_diff")
+
+        # Assert that unchanged directories remain so
+        for unchanged in unchanged_dirs:
+            self.check_populated_dir_files(unchanged)
+
+        # Assert changed directories can no longer be accessed under old names
+        for changed_dirname in changed_dirs:
+            with self.assertRaisesHTTPError(404):
+                cm.get(changed_dirname)
+
+            new_dirname = changed_dirname.replace("foo/bar", "foo/bar_diff", 1)
+
+            self.check_populated_dir_files(new_dirname)
+
+        # Created a notebook in the renamed directory should work
+        cm.new_untitled("foo/bar_diff", ext=".ipynb")
+
     def test_delete_root(self):
         cm = self.contents_manager
         with self.assertRaises(HTTPError) as err:
@@ -528,3 +651,17 @@ class TestContentsManager(TestCase):
         cm.mark_trusted_cells(nb, path)
         cm.check_and_sign(nb, path)
         assert cm.notary.check_signature(nb)
+
+
+class TestContentsManagerNoAtomic(TestContentsManager):
+    """
+    Make same test in no atomic case than in atomic case, using inheritance
+    """
+
+    def setUp(self):
+        self._temp_dir = TemporaryDirectory()
+        self.td = self._temp_dir.name
+        self.contents_manager = FileContentsManager(
+            root_dir = self.td,
+        )
+        self.contents_manager.use_atomic_writing = False
